#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2023-2024 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

from pathlib import Path
from typing import Any, Generator


class FoundSymbolInfo:
    file: File
    symbol: Symbol
    offset: int

    def __init__(self, file: File, symbol: Symbol, offset: int=0): ...

    def getAsStr(self) -> str: ...
    def getAsStrPlusOffset(self, symName: str|None=None) -> str: ...


class SymbolComparisonInfo:
    symbol: Symbol
    buildAddress: int
    buildFile: File|None
    expectedAddress: int
    expectedFile: File|None

    def __init__(self, symbol: Symbol, buildAddress: int, buildFile: File|None, expectedAddress: int, expectedFile: File|None): ...

    def diff(self) -> int|None: ...

class MapsComparisonInfo:
    badFiles: set[File] = set()
    missingFiles: set[File] = set()
    comparedList: list[SymbolComparisonInfo] = []

    def __init__(self): ...


class ProgressStats:
    undecompedSize: int
    decompedSize: int

    def __init__(self): ...

    @property
    def total(self) -> int: ...

    def undecompedPercentage(self) -> float:
        ...

    def decompedPercentage(self) -> float:
        ...

    def undecompedPercentageTotal(self, totalStats: ProgressStats) -> float:
        ...

    def decompedPercentageTotal(self, totalStats: ProgressStats) -> float:
        ...

    def getAsFrogressEntry(self, name: str) -> dict[str, int]: ...

    @staticmethod
    def getHeaderAsStr(categoryColumnSize: int=28) -> str:
        ...

    @staticmethod
    def printHeader(categoryColumnSize: int=28) -> None: ...

    def getEntryAsStr(self, category: str, totalStats: ProgressStats, categoryColumnSize: int=28) -> str:
        ...

    def print(self, category: str, totalStats: ProgressStats, categoryColumnSize: int=28) -> None: ...


class Symbol:
    name: str
    vram: int
    size: int|None # in bytes
    vrom: int|None
    align: int|None

    def __init__(self, name: str, vram: int, size: int|None=None, vrom: int|None=None, align: int|None=None): ...

    def getVramStr(self) -> str: ...
    def getSizeStr(self) -> str: ...
    def getVromStr(self) -> str: ...
    def getAlignStr(self) -> str: ...

    def serializeName(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeVram(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeSize(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeVrom(self, humanReadable: bool=True) -> str|int|None: ...

    @staticmethod
    def toCsvHeader() -> str: ...
    def toCsv(self) -> str: ...

    @staticmethod
    def printCsvHeader(): ...
    def printAsCsv(self): ...

    def toJson(self, humanReadable: bool=True) -> dict[str, Any]: ...

    def __eq__(self, other) -> bool: ...
    def __hash__(self): ...
    def __repr__(self): ...


class File:
    filepath: Path
    vram: int
    size: int # in bytes
    sectionType: str
    vrom: int|None
    align: int|None
    # _symbols: list[Symbol]

    def __init__(self, filepath: Path, vram: int, size: int, section_type: str, vrom: int|None=None, align: int|None=None) -> None: ...

    @property
    def isNoloadSection(self) -> bool: ...

    def serializeVram(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeSize(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeVrom(self, humanReadable: bool=True) -> str|int|None: ...

    #! @deprecated
    def getName(self) -> Path: ...

    def findSymbolByName(self, symName: str) -> Symbol|None: ...
    #! @deprecated: Use either `findSymbolByVram` or `findSymbolByVrom` instead.
    def findSymbolByVramOrVrom(self, address: int) -> tuple[Symbol, int]|None: ...
    def findSymbolByVram(self, address: int) -> tuple[Symbol, int]|None: ...
    def findSymbolByVrom(self, address: int) -> tuple[Symbol, int]|None: ...

    @staticmethod
    def toCsvHeader(printVram: bool=True) -> str: ...
    def toCsv(self, printVram: bool=True) -> str: ...

    @staticmethod
    def printCsvHeader(printVram: bool=True): ...
    def printAsCsv(self, printVram: bool=True): ...

    def toJson(self, humanReadable: bool=True) -> dict[str, Any]: ...

    def copySymbolList(self) -> list[Symbol]:
        """Returns a copy (not a reference) of the internal symbol list"""

    def setSymbolList(self, newList: list[Symbol]) -> None:
        """Replaces the internal symbol list with a copy of `newList`"""

    def appendSymbol(self, sym: Symbol) -> None:
        """Appends a copy of `sym` into the internal symbol list"""

    def __iter__(self) -> Generator[Symbol, None, None]: ...

    def __getitem__(self, index) -> Symbol: ...
    def __setitem__(self, index, sym: Symbol): ...
    def __len__(self) -> int: ...
    def __eq__(self, other) -> bool: ...
    def __hash__(self): ...


class Segment:
    name: str
    vram: int
    size: int
    vrom: int
    align: int|None
    # _filesList: list[File] = dataclasses.field(default_factory=list)

    def __init__(self, name: str, vram: int, size: int, vrom: int, align: int|None=None): ...

    def serializeVram(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeSize(self, humanReadable: bool=True) -> str|int|None: ...
    def serializeVrom(self, humanReadable: bool=True) -> str|int|None: ...

    def filterBySectionType(self, sectionType: str) -> Segment: ...
    def getEveryFileExceptSectionType(self, sectionType: str) -> Segment: ...

    def findSymbolByName(self, symName: str) -> FoundSymbolInfo|None: ...
    #! @deprecated: Use either `findSymbolByVram` or `findSymbolByVrom` instead.
    def findSymbolByVramOrVrom(self, address: int) -> FoundSymbolInfo|None: ...
    def findSymbolByVram(self, address: int) -> tuple[FoundSymbolInfo|None, list[File]]: ...
    def findSymbolByVrom(self, address: int) -> tuple[FoundSymbolInfo|None, list[File]]: ...

    def mixFolders(self) -> Segment: ...

    def printAsCsv(self, printVram: bool=True, skipWithoutSymbols: bool=True): ...
    def printSymbolsCsv(self): ...

    def toCsv(self, printVram: bool=True, skipWithoutSymbols: bool=True) -> str: ...
    def toCsvSymbols(self) -> str: ...

    def toJson(self, humanReadable: bool=True) -> dict[str, Any]: ...

    def copyFileList(self) -> list[File]:
        """Returns a copy (not a reference) of the internal file list"""

    def setFileList(self, newList: list[File]) -> None:
        """Replaces the internal file list with a copy of `newList`"""

    def appendFile(self, file: File) -> None:
        """Appends a copy of `file` into the internal file list"""

    def __iter__(self) -> Generator[File, None, None]: ...
    def __getitem__(self, index) -> File: ...
    def __setitem__(self, index, file: File): ...
    def __len__(self) -> int: ...
    def __eq__(self, other) -> bool: ...
    def __hash__(self): ...


class MapFile:
    def __init__(self) -> None: ...

    @staticmethod
    def newFromMapFile(mapPath: Path) -> MapFile: ...

    def readMapFile(self, mapPath: Path) -> None: ...
    def parseMapContents(self, mapContents: str) -> None: ...
    def parseMapContentsGNU(self, mapContents: str) -> None: ...
    def parseMapContentsLLD(self, mapContents: str) -> None: ...

    def filterBySectionType(self, sectionType: str) -> MapFile: ...
    def getEveryFileExceptSectionType(self, sectionType: str) -> MapFile: ...

    def findSymbolByName(self, symName: str) -> FoundSymbolInfo|None: ...
    #! @deprecated: Use either `findSymbolByVram` or `findSymbolByVrom` instead.
    def findSymbolByVramOrVrom(self, address: int) -> FoundSymbolInfo|None: ...
    def findSymbolByVram(self, address: int) -> tuple[FoundSymbolInfo|None, list[File]]: ...
    def findSymbolByVrom(self, address: int) -> tuple[FoundSymbolInfo|None, list[File]]: ...

    def findLowestDifferingSymbol(self, otherMapFile: MapFile) -> tuple[Symbol, File, Symbol|None]|None: ...

    def mixFolders(self) -> MapFile: ...

    def getProgress(self, asmPath: Path, nonmatchings: Path, aliases: dict[str, str]=dict(), pathIndex: int=2) -> tuple[ProgressStats, dict[str, ProgressStats]]: ...

    # Useful for finding bss reorders
    def compareFilesAndSymbols(self, otherMapFile: MapFile, *, checkOtherOnSelf: bool=True) -> MapsComparisonInfo: ...

    def toCsv(self, printVram: bool=True, skipWithoutSymbols: bool=True) -> str: ...
    def toCsvSymbols(self) -> str: ...

    def printAsCsv(self, printVram: bool=True, skipWithoutSymbols: bool=True): ...
    def printSymbolsCsv(self): ...

    def toJson(self, humanReadable: bool=True) -> dict[str, Any]: ...

    def copySegmentList(self) -> list[Segment]:
        """Returns a copy (not a reference) of the internal segment list"""

    def setSegmentList(self, newList: list[Segment]) -> None:
        """Replaces the internal segment list with a copy of `newList`"""

    def appendSegment(self, segment: Segment) -> None:
        """Appends a copy of `segment` into the internal segment list"""

    def __iter__(self) -> Generator[Segment, None, None]: ...
    def __getitem__(self, index) -> Segment: ...
    def __setitem__(self, index, segment: Segment): ...
    def __len__(self) -> int: ...
